/******************************************************************
*
* Software Systems Research Laboratory
* Department of Electrical Engineering and Computer Science
* University of Michigan at Ann Arbor
* 1301 Beal Avenue, Ann Arbor, MI 48105
* 
*
* Copyright (c) 1999 The Regents of The University of Michigan.
* All rights reserved.
* 
* Contact: pdmcdan@eecs.umich.edu
*******************************************************************/

//////////////////////////////////////////////////////////////////////////////////
//
//
//   File : apcc.cpp
//
//   Created : 14-Aug-00
//   By : Patrick McDaniel
//
//   Description : This is the demonstation application for the
//                 Antigone 2.0 system.
//
//   *** Change History ***
//
//   14-Aug-00 Created File : DPM
//
//////////////////////////////////////////////////////////////////////////////////

// Include Files

// Project Include Files
#include <AGroup.h>
#include <Environment.h>

// Defines
#define USAGE \
"USAGE aapp <mc/srv addr> <port> <loc policy> [-p=group policy] [-pl=<policy list>]" \
" [-aa=<sa>]  [-ap=<sp>] [-l=<log>] [-w=<lifetime>] -o -t\n" \
"\n" \
"where " \
"\n" \
"    mc/srv addres - is multicast (ASYM/SYMMETRIC multicast) \n" \
"                    or server (PT2PT modes)\n" \
"    port          - port associated with mc/srv address\n" \
"    loc policy    - file containing local policy\n" \
"    group policy  - file containing group policy (server only)\n" \
"    policy list   - file containing local policy file list (server only)\n" \
"    sa            - server address (ASYMMETIC multicast)\n" \
"    sp            - server port (ASYMMETIC multicast)\n" \
"    log           - filename for application log\n" \
"    lifetime      - lifetime of application (in seconds)\n" \
"    t             - timestamp logs\n" \
"    o             - enable mcast loopback\n"

// AAPP Implementor
class AppImplementor : public APolicyImplementor {

  public: // Public Interface

    //
    // Constructors and Destructors

    // Attribute Constructor
    AppImplementor( AuditLog * lg )
      : log( lg )
    {
    }

    // Destructor
    virtual ~AppImplementor( void )
    {
    }

    //
    // Implementor Methods

    // Evaluate a application conditional
    virtual Boolean EvaluateCondition( ACondEval & cond )
    {
        // Get the conditional predicate
        String pred = APolicyImplementor::Predicate( cond );
        if ( strcasecmp((char *)pred, "hostname") == 0 )
        {
            // Check argumements, get test hostname
            String tst, hname;
            if ( (APolicyImplementor::NoArgs(cond) != 1) || 
                 (APolicyImplementor::Arg(0,cond,tst) == FALSE) )
            {
                // Error out
                LOGMSG1( log, AGRP_ALEVEL, "Malformed hostname condition [%s]", 
                         (char *)cond );
                return( FALSE );
            }

            // Test, set evaluation result information
            cond.ResultType() = ACondEval::APDL_INVARIANT;
            cond.Result() = (tst == IPAddress::GetLocalHostname(hname)) 
                             ? TRUE : FALSE;
            LOGMSG3( log, AGRP_ALEVEL, "%s Evaluate (hname = %s) INVARIENT %s",
                     (char *)pred, (char *)hname, BOOLE_STR(cond.Result()) );

            // Return evaluated
            return( TRUE );
        }

        // Return not implemented by this object
        return( FALSE );
    }

  private: // Private Interface and Data

    // 
    // Class Attributes

    AuditLog * log;
      // This is the log to which to write debug information

};

#define PE_DEFAULT_ALEVEL DEFAULT_ALEVEL|AMECH_ALEVEL|AGRP_ALEVEL|APDL_ALEVEL|SSL_ALEVEL
#define SEND_PERIOD 500
#define WAIT_INDEF 100000
#define EJECT_ONCE

// Static Application Data
AuditLog *alog = NULL;

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : executeGroup
// Description  : server operation function
//
// Inputs       : grp - group 
// Outputs      : 0 if successful, -1 if failure

Boolean executeGroup( AGroup *grp )
{
    // Stop on exceptions
    try
    {
        // Log the local and group policies, connect to the group
        if ( grp->IsServer() == TRUE )
            grp->Policy()->logPolicy( APDL_ALEVEL, alog );
        if ( grp->Connect() != TRUE )
        {
            // Error out
            LOGMSG0( alog, FATAL_ERROR, "Error connecting/joining group." );
            return( FALSE );
        }

        // Get session lifetime application policy
        unsigned long ltime = WAIT_INDEF;
        String val, ejectee;
        if ( grp->AttributeSet().getAttributeValue("lifetime", val) == TRUE ) 
        {
            // Get, log policy
            sscanf( (char *)val, "%ld", &ltime );
            LOGMSG1( alog, PE_DEFAULT_ALEVEL, "Application policy (lifetime"
                     " = %d)", ltime );
        }

        // Setup the session lifetime/ejection timers
        AtkTimer ltimer( ltime*1000 ), *ejectTimer = NULL;
        ltimer.reset();

        // Get ejection policy, if any
        if ( grp->AttributeSet().getAttributeValue("eject", ejectee) == TRUE ) 
        {
            // Get eject policy
            ejectTimer = new AtkTimer( WAIT_INDEF*1000 );
            ejectTimer->reset();
            LOGMSG1( alog, PE_DEFAULT_ALEVEL, "Application Policy (eject %s)", 
                     (char *)ejectee );
        }

        // Get session lifetime application policy
        unsigned long speriod = SEND_PERIOD;
        if ( grp->AttributeSet().getAttributeValue("sendperiod", val) == TRUE ) 
        {
            // Get, log policy
            sscanf( (char *)val, "%ld", &speriod );
            LOGMSG1( alog, PE_DEFAULT_ALEVEL, "Application policy (sendperiod"
                     " = %d)", speriod );
        }

        // Transmit periodically
        String snder;
        unsigned long idx = 1, rindx;
        AtkTimer timer( speriod );
        timer.reset();
        while ( (grp->State() == AGroup::AG_INOPERATION) && 
                (ltimer.expired() == FALSE) )
        {
            // Keep reading messages
            Buffer *buf = NULL;
            if ( (grp->readMessage( &buf, &timer ) == TRUE) && (buf != NULL) )
            {
                // Read message, log it
              //(*buf) >> snder;
              //(*buf) >> rindx;
              //LOGMSG2( alog, PE_DEFAULT_ALEVEL, "App RECV [%s,%lx]",
              //         (char *)snder, rindx );
              LOGMSG0( alog, PE_DEFAULT_ALEVEL, "App RECV");
                delete buf;
            }

            // If ejecting and expired, then do it
            if ( (grp->IsServer() == FALSE) && (ejectTimer != NULL) && 
                 (ejectTimer->expired()) )
            {
                grp->Eject( ejectee );
                delete ejectTimer;
                ejectTimer = NULL;
            }

            /* If timer expired, then send one
               if ( timer.expired() )
               {
               // Setup buffer, send it
               timer.reset();
               Buffer *buf = new Buffer();
               (*buf) << grp->LocalID();
               (*buf) << idx ++;
               grp->sendMessage( buf );
               }*/
        }

        // Quit the group
        grp->Quit();

        // Cleanup
        delete ejectTimer;
    }
    catch (AtkException * e)
    {
        // Print out exception, error out
        LOGMSG1( alog, FATAL_ERROR, "Exception [%s]", e->Description() );
        cerr << (char *)e->toString() << ", aborting." << endl;
        return( FALSE );
    }

    // Return sucessfully
    return( TRUE );
}

//////////////////////////////////////////////////////////////////////////////////
//
// Method       : main
// Description  : main function
//
// Inputs       : argc - input arguments
//                argv - pointers to input arguments
// Outputs      : 0 if successful, -1 if failure

int main( int argc, char **argv )
{
    // Create some definitions for CLP and ENV
    const char * clps[] = { "!<1>=s", "!<2>=i", "!<3>=s", "p=s", "aa=s", 
                      "ap=i", "l=s", "t=b", "o=b", "pl=s" };
    char *usage = USAGE; 

    // Always check for exception in processing
    Environment *env;
    String alogs, gaddr, lpol, aaddr, plist;
    int gport, aport;
    ATransport *trans;
    ATransport::AtMode mode = ATransport::AT_UNKNOWN;
    AGroup *group;
    Boolean isServer = FALSE;
    try
    {
        // Create the environment, get arguments
        env = new Environment( 10, (char**)clps, 0, NULL, argc, argv, usage );

        // Check environment

        // Set timestamping on/off
        Boolean ts = env->clpHasValue("t");

        // Create log (with or without file echo)
        if ( env->clpHasValue("l") )
        {
            // Create log with file echo
            alogs = env->getStringCLPValue("l");
            alog = new AuditLog( alogs, TRUE, TRUE, ts );
        }
        else
            // Create log without file echo
            alog = new AuditLog( NULL, TRUE, FALSE, ts );
        alog->setAuditLevel( PE_DEFAULT_ALEVEL );

        // Get the group address (or server address)
        gaddr = env->getStringCLPValue("<1>");
        gport = env->getIntCLPValue("<2>");
        lpol = env->getStringCLPValue("<3>");
        isServer = env->clpHasValue("p");

        // Get the policy list 
        if ( env->clpHasValue("pl") ) 
            plist = env->getStringCLPValue("pl");
        else
            plist = "";

        // Set the multicast loopback policy (if specified)
        if ( env->clpHasValue("o") ) ATransport::SetLoopback( TRUE );

        // First, setup the transport layer address
        IPAddress *gipaddr = IPAddress::IPAddressFactory( gaddr, gport ), 
                  *saddr = NULL;
        if ( gipaddr == NULL )
        {
            // Error out
            LOGMSG1( alog, FATAL_ERROR, "Unable to interpret group address [%s]", 
                     (char *)gaddr );
            return( -1 );
        }

        // If not multicast address, then we are setting up a server address
        if ( gipaddr->IsMulticastAddress() == TRUE )
        {
            // Are we are setting asymmetric address
            if ( env->clpHasValue("ap") )
            {
                // Asymmetric multicast, set it up
                aport = env->getIntCLPValue("ap");
                aaddr = (env->clpHasValue("aa")) ? env->getStringCLPValue("aa") : 
                        (String)A_LOCALHOST;
                if ( (saddr = IPAddress::IPAddressFactory(aaddr,aport)) == NULL )
                {
                    // Error out
                    LOGMSG1( alog, FATAL_ERROR, "Unable to interpret asymmetric"
                             " address [%s]", (char *)gaddr );
                    return( -1 );
                }

                // Check if client has defined server address
                if ( (isServer == FALSE) && ((env->clpHasValue("aa") == FALSE)) )
                {
                    // Error out
                    LOGMSG0( alog, FATAL_ERROR, "Undefined server asymmetric"
                             " address [-aa=??]" );
                    return( -1 );
                }
                
                // Create assymetric transport object
                mode = ATransport::AT_ASYMMETRIC;
            }
            else
                // Set to symmetric mode
                mode = ATransport::AT_SYMMETRIC;
        }
        else
          // Set to point to point mode
          mode = ATransport::AT_PTTOPT;

        // Now setup an attribute table
        AAttributeSet attrs;
        StringVar sv( "grouptype", "aapp" );
        AAttribute *gtype = new AAttribute( &sv );
        attrs.addAttribute( gtype );

        // Now create the transport, then group
        AppImplementor impl( alog );
        if ( isServer == TRUE )
        {
            // Server, use group or asym port, then create group
            trans = new ATransport( gipaddr, (saddr == NULL) ? gipaddr->Port() : 
                                    saddr->Port(), mode );
            trans->Log() = alog;
            group = new AGroup( trans, env->getStringCLPValue("p"), 
                                lpol, plist, &impl, &attrs, alog );
        }
        else
        {
            // Client, use group or asym addresss, then create group
            if ( mode == ATransport::AT_PTTOPT )
                // unicast address is used as (virtual) mcast address on P2P
                trans = new ATransport( (IPAddress *)NULL, gipaddr, mode );
            else
                trans = new ATransport( gipaddr, saddr, mode );
            trans->Log() = alog;
            group = new AGroup( trans, lpol, &impl, &attrs, alog );
        }

        // Do the group context
        executeGroup( group );
    }
    catch (AtkException * e)
    {
        // Print out exception, error out
        LOGMSG1( alog, FATAL_ERROR, "Exception [%s]", e->Description() );
        cerr << (char *)e->toString() << ", aborting." << endl;
        cerr << usage << endl;
        return( -1 );
    }

    // Cleanup 
    delete alog;
    delete env;
    delete group;
    delete trans;

    // Exit success/failure return code
    return( 0 );
}
